-- ForthEC ARM code optimiser 
-- /Mic, 2004


include parser.e


-- ARM registers
constant regs32 = {"r0","r1","r2","r3","r4","r5","r6","r7","r8"}

-- ARM condition codes ("nz" is left out because of how these are used in the patterns below)
constant conds = {"g","ge","l","le","z"}



-- Instruction patterns to scan for
--
--
constant intpatterns = {
{
	{
	"ldr r1,[r10]",
	"str r0,[r10,#-4]!",
	"mov r0,r1",
	"ldr r1,[r10],#4"
	},
	{
	{"ldr r2,[r10]"},
	{"mov r1,r0"},
	{"mov r0,r2"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	{"add r0,r0,r1","and r0,r0,r1","orr r0,r0,r1","eor r0,r0,r1"}
	},
	{
	{4,1,3," r0,r0,#$1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	"rsb r0,r0,r1"
	},
	{
	{"sub r0,r0,#$1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	"cmp r1,r0"
	},
	{
	--{"mov r1,#$1"},
	{"cmp r0,#$1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"cmp r0,#/imm/",
	"ldr r0,[r10],#4",
	"bne /label/"
	},
	{
	{"ldr r1,=$1"},
	{"cmp r1,#$2"},
	5
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"cmp r0,#/imm/",
	"ldr r0,[r10],#4",
	"bne /label/"
	},
	{
	{"ldr r1,=$1"},
	{"cmp r1,#$2"},
	5
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,=/const/",
	"str r0,[r1]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r2,=$1"},
	3,
	{"str r2,[r1]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,r1",
	"cmp r0,#/imm/",
	"ldr r0,[r10],#4"
	},
	{
	{"cmp r1,#$1"}
	}
},
{
	{
	"ldr r1,[r10]",
	"str r0,[r10]",
	"mov r0,r1",
	{"add r0,r0,#/imm/","sub r0,r0,#/imm/","and r0,r0,#/imm/","orr r0,r0,#/imm/"},
	"ldr r1,[r10]",
	"str r0,[r10]",
	"mov r0,r1"
	},
	{
	1,
	{4,1,3," r1,r1,#$1"},
	{"str r1,[r10]"}
	}
},
{
	{
	"ldr r2,[r10]",
	"mov r1,r0",
	"mov r0,r2",
	{"mul r0,r1,r0","add r0,r0,r1","and r0,r0,r1"}
	},
	{
	{"ldr r1,[r10]"},
	4
	}
},
{
	{
	{"and r0,r0,#/imm/","add r0,r0,#/imm/","orr r0,r0,#/imm/","eor r0,r0,#/imm/"},
	"cmp r0,#0"
	},
	{
	{1,1,3,"s r0,r0,#$1"}
	}
},
{
	{
	"moveq r0,#0",
	"movne r0,#-1",
	"cmp r0,#-1",
	"ldr r0,[r10],#4",
	"bne /label/"
	},
	{
	4,
	{"beq $1"}
	}
},
{
	{
	"moveq r0,#-1",
	"movne r0,#0",
	"cmp r0,#-1",
	"ldr r0,[r10],#4",
	"bne /label/"
	},
	{
	4,
	{"bne $1"}
	}
},
{
	{
	"beq /label/",
	"b /label/",
	"/label/"
	},
	{
	{"bne $2"},
	3
	}
},
{
	{
	"bne /label/",
	"b /label/",
	"/label/"
	},
	{
	{"beq $2"},
	3
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"str r7,[r12,#-4]!",
	"str r6,[r12,#-4]!",
	"ldr r7,[r10],#4",
	"mov r6,r0",
	"ldr r0,[r10],#4"
	},
	{
	5,
	6,
	{"ldr r7,=$1"},
	{"mov r6,#$2"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	{"ldrh r0,[r0]","ldrb r0,[r0]"},
	{"ands r0,r0,#/imm/","orrs r0,r0,#/imm/","eors r0,r0,#/imm/"},
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r2,=$1"},
	{3,1,4," r1,[r2]"},
	{4,1,4," r1,r1,#$2"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,[r10],#4",
	"str r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,=$1"},
	{"ldr r2,=$2"},
	{"str r1,[r2]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,[r10],#4",
	{"strb r1,[r0]","strh r1,[r0]"},
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,=$1"},
	{"ldr r2,=$2"},
	{6,1,4," r1,[r2]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,[r10],#4",
	"str r0,[r1],#4",
	"mov r0,r1"
	},
	{
	{"ldr r1,=$1"},
	{"str r1,[r0],#4"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	"mov r0,r1,lsl r0"
	},
	{
	{"mov r0,r0,lsl#$1"}
	}
},
{
	{
	"mov r0,/reg32/",
	"mov r0,r0,lsl#/imm/"
	},
	{
	{"mov r0,$1,lsl#$2"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"str r7,[r12,#-4]!",
	"str r6,[r12,#-4]!",
	"ldr r7,[r10],#4",
	"mov r6,r0",
	"ldr r0,[r10],#4"
	},
	{
	5,
	6,
	{"mov r6,#$2"},
	{"mov r7,#$1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r2,[r10]",
	"mov r1,r0",
	"mov r0,r2",
	"strh r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,=$1"},
	{"strh r1,[r0]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	"mov r0,r1,lsr r0"
	},
	{
	{"mov r0,r0,lsr#$1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,[r10],#4",
	"add r0,r0,r1"
	},
	{
	{"ldr r1,=$1"},
	{"add r0,r0,r1"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"cmp r0,#/imm/",
	"ldr r0,[r10],#4"
	},
	{2}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r0,[r10,r0,lsl#2]",
	"ldr r1,[r10],#4"
	},
	{
	{"mov r1,r0"},
	{"ldr r0,[r10,#%($1-#1*#4)]"}
	}
},
{
	{
	"mov r0,#/imm/",
	"ldr r0,[r10,r0,lsl#2]"
	},
	{
	{"ldr r0,[r10,#%($1*#4)]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,r0,lsl#/imm/",
	"ldr r1,[r10],#4",
	"orr r0,r0,r1"
	},
	{
	{"orr r0,r0,r0,lsl#$1"}
	}
},
{
	{
	"add r0,r0,#/imm/",
	{"ldrh r0,[r0]","ldrb r0,[r0]"}
	},
	{
	{2,1,4," r0,[r0,#$1]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r1,r0",
	"ldr r0,[r10,#/imm/]",
	"strh r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,[r10,#%($1-#4)]"},
	{"strh r0,[r1]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,[r10,#/imm/]",
	"add r0,r0,#/imm/",
	"ldr r1,[r10],#4",
	"strh r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,[r10,#%($1-#4)]"},
	{"strh r0,[r1,#$2]"},
	{"ldr r0,[r10],#4"}
	}
},
{
	{
	"ldr r1,[r10]",
	"str r0,[r10]",
	"mov r0,r1",
	"ldr r1,=/imm/",
	"add r0,r0,r1",
	"ldr r1,[r10]",
	"str r0,[r10]",
	"mov r0,r1",
	"ldr r1,=$1",
	"add r0,r0,r1"
	},
	{
	4,
	{"ldr r2,[r10]"},
	{"add r0,r0,r1"},
	{"add r2,r2,r1"},
	{"str r2,[r10]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"add r6,r6,r0",
	"ldr r0,[r10],#4"
	},
	{
	{"add r6,r6,#$1"}
	}
},
{
	{
	"ldr r0,[r10],#4",
	"ldr r0,[r10],#4"
	},
	{
	{"add r10,r10,#4"},
	2
	}
},
{
	{
	"ldr r1,[r10]",
	"str r0,[r10,#-4]!",
	"mov r0,r1",
	"add r0,r0,#/imm/",
	"ldr r1,[r10],#4",
	"strh r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,[r10],#4"},
	{"strh r0,[r1,#$1]"},
	{"mov r0,r1"}
	}
},

{
	{
	"beq /label/",
	{"sub r0,r0,#/imm/","add r0,r0,#/imm/"},
	"$1"
	},
	{
	{2,1,3,"ne r0,r0,#$2"}
	}
},
{
	{
	"ldr r0,=/imm/",
	"ldr r1,=/imm/",
	"add r0,r0,r1"
	},
	{
	{"ldr r0,=%($1+$2)"}
	}
},
{
	{
	"mov r0,r1",
	"ldrh r0,[r0,#/imm/]"
	},
	{
	{"ldrh r0,[r1,#$1]"}
	}
},
{
	{
	"ldr r1,[r10]",
	"str r0,[r10]",
	{"ldrh r0,[r1,#/imm/]","ldrb r0,[r1,#/imm/]"},
	{"and r0,r0,#/imm/","add r0,r0,#/imm/","orr r0,r0,#/imm/"},
	"ldr r1,[r10],#4",
	{"add r0,r0,r1","and r0,r0,r1","orr r0,r0,r1"}
	},
	{
	{"ldr r1,[r10],#4"},
	{3,1,4," r2,[r1,#$1]"},
	{4,1,3," r2,r2,#$2"},
	{6,1,3," r0,r2,r0"}
	}
},
{
	{
	"mov r0,r1",
	"ldr r1,=/imm/",
	"add r0,r0,r1",
	{"ldrh r0,[r0]","ldrb r0,[r0]"}
	},
	{
	{"ldr r2,=$1"},
	{4,1,4," r0,[r1,r2]"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"mov r0,#/imm/",
	"ldr r1,[r10],#4",
	"str r0,[r1],#4",
	"mov r0,r1"
	},
	{
	{"mov r2,#$1"},
	{"str r2,[r0],#4"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/imm/",
	"ldr r1,[r10],#4",
	"str r0,[r1],#4",
	"mov r0,r1"
	},
	{
	{"ldr r2,=$1"},
	{"str r2,[r0],#4"}
	}
},
{
	{
	"str r0,[r10,#-4]!",
	"ldr r0,=/var/",
	"ldr r0,[r0]",
	{"add r0,r0,#/imm/","sub r0,r0,#/imm/","eor r0,r0,#/imm/"},
	"str r0,[r10,#-4]!",
	"ldr r0,=$1",
	"ldr r1,[r10],#4",
	"str r1,[r0]",
	"ldr r0,[r10],#4"
	},
	{
	{"ldr r1,=$1"},
	{"ldr r2,[r1]"},
	{4,1,3," r2,r2,#$2"},
	{"str r2,[r1]"}
	}
}
}




global integer
	linesRemoved,
	constRemoved,
	optLevel

global sequence
	patvars,
	constlist
	
linesRemoved = 0
constRemoved = 0
constlist = {}


-- Is this a pop instruction ?
function is_pop(sequence s)
	if length(s)>3 then
		if equal(s[1..3],"pop") then
			return 1
		end if
	end if
	return 0
end function


-- Is this a push instruction ?
function is_push(sequence s)
	if length(s)>4 then
		if equal(s[1..4],"push") then
			return 1
		elsif length(s)>19 then
			if equal(s[1..19],"mov dword ptr [esp]") then
				return 2
			end if
		end if
	end if
	return 0
end function


function get_reg(sequence s)
	integer p
	
	p = find(',',s)
	if p then
		return s[p+1..length(s)]
	else
		return s[find(' ',s)+1..length(s)]
	end if
end function


function get_reg1(sequence s)
	integer p
	
	p = find(',',s)
	if p then
		return s[find(' ',s)+1..p-1]
	else
		return s[find(' ',s)+1..length(s)]
	end if
end function


function get_imm(sequence s)
	return s[find(',',s)+1..length(s)]
end function


-- safe_compare(code,pattern)
function safe_compare(sequence s1,sequence s2)
	if length(s1)>=length(s2) then
		return equal(s1[1..length(s2)],s2)
	end if
	return 0
end function


function compare_patterns(sequence s1,sequence s2)
	integer eql,p1,p2,m,n,o
	sequence s3,s4
	
	eql = 1
	p1 = 1
	p2 = 1
	m = length(patvars)
	o = 0
	
	
	while eql do
		if p1>length(s1) or
		   p2>length(s2) then
			exit
		end if
		if s2[p2] = '/' then
			s3 = {}
			p2 += 1
			while s2[p2] != '/' do
				s3 &= s2[p2]
				p2 += 1
			end while
			p2 += 1
			
			s4 = {}
			while p1<=length(s1) do
				if s1[p1]=',' or
				   s1[p1]=']' then
					exit
				end if
				s4 &= s1[p1]
				p1 += 1
			end while
			
			if equal(s3,"reg32") then
				if find(s4,regs32) then
					patvars = append(patvars,s4)
				else
					eql = 0
					exit
				end if
			elsif equal(s3,"reg32d") then
				if find(s4,regs32[2..length(regs32)]) then
					--puts(1,s4&"\n")
					patvars = append(patvars,s4)
				else
					eql = 0
					exit
				end if
			elsif equal(s3,"const") then
				--s4 &= ']'
				eql = 0
				if find('_',s4) and length(s4)>=6 then
					if equal("const_",s4[1..6]) then
						eql = 1
						--p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"var") then
				--s4 &= ']'
				eql = 0
				if find('_',s4) then
					if equal("var_",s4[1..4]) then
						eql = 1
						--p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"ind") then
				s4 &= ']'
				eql = 0
				if length(s4)>12 then
					if equal("dword ptr [e",s4[1..12]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"qlit") then
				s4 &= ']'
				eql = 0
				if find('_',s4) then
					if equal("qword ptr [lit_",s4[1..15]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"imm") then
				o = 0
				if length(s4)>6 then
					if equal(s4[1..6],"offset") then
						patvars = append(patvars,s4)
						o = 1
					end if
				end if
				if o=0 then
					s3 = value(s4)
					if s3[1]=GET_SUCCESS and integer(s3[2]) then
						patvars = append(patvars,s4)
					else
						eql = 0
					end if
				end if
			elsif equal(s3,"cond") then
				if find(' ',s4) then
					p1 -= length(s4)+1-find(' ',s4)
					--puts(1,"s1[p1] = "&s1[p1]&"\n")
					s4 = s4[1..find(' ',s4)-1]
				end if
				if find(s4,conds) then
					patvars = append(patvars,s4)
					--puts(1,s4&".\n")
				else
					eql = 0
				end if
			elsif equal(s3,"label") then
				if s4[1]='_' then
					patvars = append(patvars,s4)
				else
					eql = 0
				end if
			elsif equal(s3,"if") then
				if s4[1]='_' then
					patvars = append(patvars,s4)
				else
					eql = 0
				end if				
			elsif equal(s3,"mem32") then
				s4 &= ']'
				eql = 0
				if length(s4)>11 then
					if equal("dword ptr [",s4[1..11]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			elsif equal(s3,"mem") then
				s4 &= ']'
				eql = 0
				if length(s4)>11 then
					if equal("dword ptr [",s4[1..11]) or
					   equal("qword ptr [",s4[1..11]) then
						eql = 1
						p1 += 1
						patvars = append(patvars,s4)
					end if
				end if
			end if
			
		elsif s2[p2]='$' then
			n = 0
			p2 += 1
			while p2<=length(s2) do
				if s2[p2]>='0' and s2[p2]<='9' then
					n = n*10 + s2[p2]-'0'
				else
					exit
				end if
				p2 += 1
			end while
			s3 = patvars[n]

			s4 = {}
			while p1<=length(s1) do
				if s1[p1]=',' or
				   s1[p1]=']' then
				   	if s1[p1]=']' then p1+=1 end if
					exit
				end if
				s4 &= s1[p1]
				p1 += 1
			end while
			if length(s4)>11 then
				if equal(s4[7..9],"ptr") then
					s4 &= ']'
				end if
			end if
			
			if not equal(s3,s4) then
				if s3[1]='_' then
					if not equal(s3&':',s4) then
						eql = 0
					end if
				else
					eql = 0
				end if
			end if
				
		else
			if s1[p1]!=s2[p2] then
				eql = 0
			end if
			p1 += 1
			p2 += 1
		end if
	end while
	
	if p1<length(s1) or p2<length(s2) then
		eql = 0
	end if
	
	if not eql then
		if m=0 then
			patvars = {}
		else
			patvars = patvars[1..m]
		end if
	end if
	
	return eql
end function



function pattern_append(sequence s1,sequence s2)
	integer p2,m,n,oper,expval
	sequence s3,s4,ops

	p2 = 1
	s3 = {}
	while 1 do
		if p2>length(s2) then
			exit
		end if
		
		if s2[p2]='$' then
			n = 0
			p2 += 1
			while p2<=length(s2) do
				if s2[p2]>='0' and s2[p2]<='9' then
					n = n*10 + s2[p2]-'0'
				else
					exit
				end if
				p2 += 1
			end while
			s3 &= patvars[n]
		elsif s2[p2]='%' then
			p2 += 2
			expval = 0
			ops = {}
			oper = 0
			while s2[p2] != ')' do
				if s2[p2] = '$' or
				   s2[p2] = '#' then
				   	m = s2[p2]
					n = 0
					p2 += 1
					while p2<=length(s2) do
						if s2[p2]>='0' and s2[p2]<='9' then
							n = n*10 + s2[p2]-'0'
						else
							exit
						end if
						p2 += 1
					end while
					if m='$' then
						s4 = value(patvars[n])
					else
						s4 = {GET_SUCCESS,n}
					end if
					if oper=0 then
						expval = s4[2]
					elsif oper='+' then
						expval += s4[2]
					elsif oper='-' then
						expval -= s4[2]
					elsif oper='*' then
						expval *= s4[2]
					end if

				elsif s2[p2] = '+' then
					oper = s2[p2]
					p2 += 1
				elsif s2[p2] = '*' then
					oper = s2[p2]
					p2 += 1
				elsif s2[p2] = '-' then
					oper = s2[p2]
					p2 += 1
				else
					p2 += 1
				end if
			end while
			p2 += 1
			s3 &= sprintf("%d",expval)
					
		else
			s3 &= s2[p2]
			p2 += 1
		end if
	end while
	
	return append(s1,s3)
end function




global function optimise(sequence subject,integer remConst)
	integer i1,i2,i3,i4,p,q,n,m,o,clean,improvement,times,reachable
	sequence s,r,t,u,pat

	if remConst and 0 then
		-- Remove unused constants
		for i=1 to length(constants[2]) do
			if constants[2][i][2]=0 then
				subject = subject[1..constants[2][i][3]-1] & {"add esp,4"} & subject[constants[2][i][3]+2..length(subject)]
				constRemoved += 1
				for j=i+1 to length(constants[2]) do
					if constants[2][j][3]>constants[2][i][3] then
						constants[2][j][3] -= 1
					end if
				end for
			end if
		end for
	end if


	if optLevel >= 5 and 0 then
		if remConst then
			constlist = {}
			for i=1 to length(subject) do
				patvars = {}
				if compare_patterns(subject[i],"ldr r1,/const/") then
					if equal(subject[i+1],"str r0,[r1]") then
						constlist = append(constlist,patvars&{i,1,1})
					elsif equal(subject[i+1],"str r2,[r1]") then
						constlist = append(constlist,patvars&{i,2,1})
					end if
				end if
			end for
		end if

		--patvars = {}
		--for i=1 to length(constlist) do
		--	patvars = constlist[i][1]
		--	for j=1 to length(subject) do
		--		if equal(subject[j],"push "&patvars[1]) then
		--			subject[j] = "pushdw "&patvars[2]
		--			constlist[i][4] = 0
		--		elsif compare_patterns(subject[j][4..length(subject[j])]," /reg32/,$1") then
		--			subject[j] = subject[j][1..find(',',subject[j])-1]&","&patvars[2]
		--			patvars = patvars[1..2]
		--			constlist[i][4] = 0
		--		end if
		--	end for
		--end for

		--p=2
		--while 1 do
		--	if p+2>length(subject) then
		--		exit
		--	end if
		--	if equal(subject[p],"sub esp,4") then
		--		patvars = {}
		--		if compare_patterns(subject[p+1],"mov dword ptr [esp],/imm/") then
		--			subject = subject[1..p-1] & {"pushdw "&patvars[1]} & subject[p+2..length(subject)]
		--			linesRemoved += 1
		--		end if
		--	end if
		--	p += 1
		--end while
	end if		
	
	
	if optLevel >= 5 then
		-- Run up to 10 passes
		times = 10
		while times>0 do
			improvement = linesRemoved
			
			-- Replace common instruction patterns by more compact forms
			p = 1
			while p<=length(subject) do
				for i=1 to length(intpatterns) do
					pat = intpatterns[i]
					if p+length(pat[1])<=length(subject)+1 then
						patvars = {}
						n = 1
						for j=1 to length(pat[1]) do

							if sequence(pat[1][j][1]) then
								m = 0
								for k=1 to length(pat[1][j]) do
									if compare_patterns(subject[p+j-1],pat[1][j][k]) then
										m = k
										exit
									end if
								end for
								if not m then
									n = 0
									exit
								end if
							else
								if not compare_patterns(subject[p+j-1],pat[1][j]) then
									n = 0
									exit
								end if
							end if
						end for

						-- Was a matching pattern found ?
						if n then
							if i=5 or i=6 then
								if equal(patvars[1],patvars[2]) then
									linesRemoved += 5
									subject = subject[1..p-1] & subject[p+5..length(subject)]
								else
									linesRemoved += 4
									subject = subject[1..p-1] & {"b "&patvars[3]} & subject[p+5..length(subject)]
								end if
							else
								s = {}
								for j=1 to length(pat[2]) do
									if sequence(pat[2][j]) then
										if sequence(pat[2][j][1]) then
											s = pattern_append(s,pat[2][j][1])
										else
											s = pattern_append(s,subject[p+pat[2][j][1]-1][pat[2][j][2]..pat[2][j][3]]&pat[2][j][4])
										end if
									else
										s = pattern_append(s,subject[p+pat[2][j]-1])
									end if
								end for
								subject = subject[1..p-1] & s & subject[p+length(pat[1])..length(subject)]
								linesRemoved += length(pat[1])-length(pat[2])
							end if
							exit
						end if
					end if
				end for
				p += 1
			end while

			
			-- Stop when there's nothing more to optimise
			improvement -= linesRemoved
			if not improvement then
				exit
			end if
			times -= 1
		end while
	end if
	
	
	return subject
end function


